# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 28.7 `blgm` Chapter.

``blgm`` - Linear Model Specification

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

The functions in this module provide a mechanism for specifying a linear
model using a text based modelling language and are intended to be used
in conjunction with the model fitting functions from other modules, for
example submodule ``correg``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/g22/g22intro.html
"""

# NAG Copyright 2017-2022.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import utils

def lm_formula(hform, formula):
    u"""
    Specify a linear model via a formula string.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``, ``optget``.

    ``lm_formula`` parses a text string containing a formula specifying
    a linear model and outputs a G22 handle to an internal data
    structure.
    This G22 handle can then be passed to various functions in submodule
    ``blgm``.
    In particular, the G22 handle can be passed to ``lm_design_matrix``
    to produce a design matrix or ``lm_submodel`` to produce a vector of
    column inclusion flags suitable for use with functions in submodule
    ``correg``.

    For full information please refer to the NAG Library document for
    g22ya

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/g22/g22yaf.html

    Parameters
    ----------
    hform : Handle, modified in place
        `On entry`: must be set to a null Handle, alternatively an
        existing G22 handle may be supplied in which case this function
        will destroy the supplied G22 handle as if ``handle_free`` had
        been called.

        `On exit`: holds a G22 handle to the internal data structure
        containing a description of the model M as specified in
        `formula`. You **must not** change the G22 handle other than
        through functions in submodule ``blgm``.

    formula : str
        A string containing the formula specifying M. See [Notes] for
        details on the allowed model syntax.

    Other Parameters
    ----------------
    'Contrast' : str
        Default = 'FIRST'

        This argument controls the default contrasts used for the
        categorical independent variables appearing in the model.
        Six types of contrasts and dummy variables are available:

        'FIRST'
            Treatment contrasts relative to the first level of the
            variable will be used.

        'LAST'
            Treatment contrasts relative to the last level of the
            variable will be used.

        'SUM FIRST'
            Sum contrasts relative to the first level of the variable
            will be used.

        'SUM LAST'
            Sum contrasts relative to the last level of the variable
            will be used.

        'HELMERT'
            Helmert contrasts will be used.

        'POLYNOMIAL'
            Polynomial contrasts will be used.

        'DUMMY'
            Dummy variables will be used rather than a contrast.

        See ``lm_design_matrix`` for more information on contrasts,
        their effect on the design matrix and how they are constructed.

        This argument may have an `instance identifier` associated with
        it (see ``optset`` and ``optget``).
        The `instance identifier` must be the name of one of the
        variables appearing in the model supplied in `formula` when the
        G22 handle was created.
        For example, `CONTRAST : VAR1 = HELMERT` would set Helmert
        contrasts for the variable named `VAR1`.

        If no `instance identifier` is specified, the default contrast
        for all categorical variables in the model is changed, otherwise
        only the default contrast for the named variable is changed.

        In some situations it might be necessary for a variable to use a
        different contrast, depending on where it appears in the model
        formula.
        In order to allow contrasts to be specified on a term by term
        basis the @ operator can be used in the model formula.
        The syntax for this operator is V_j@c, where c is one of: `F`,
        `L`, `SF`, `SL`, `H`, `P` or `D`, corresponding to treatment
        contrasts relative to the first and last levels, sum contrasts
        relative to the first and last levels, Helmert contrasts,
        polynomial contrasts or dummy variables respectively.

        If the contrast has not been explicitly specified via the @
        operator, the value obtained from the option 'Contrast' is used.

        For example, setting `formula` to `VAR1 + VAR1@H.VAR2@P +
        VAR2@H.VAR3`, specifies that the variable named `VAR1` should
        use the default contrasts in the first term and Helmert
        contrasts in the second term.
        The variable named `VAR2` should use polynomial contrasts in the
        second term and Helmert contrasts in the third term.
        The variable named `VAR3` should use the default contrasts in
        the third term.

    'Explicit Mean' : str
        Default = 'NO'

        If 'Explicit Mean' = 'YES', any mean effect included in the
        model will be explicitly added to the design matrix, X, as a
        column of 1s.

        If 'Explicit Mean' = 'NO', it is assumed that the function to
        which X will be passed treats the mean effect as a special case,
        see mean in ``correg.linregm_fit`` for example.

    'Formula' : str
        This argument returns a verbose version of the model formula
        specified in `formula`, expanded and simplified to only contain
        variable names, the operators + and . and any contrast
        identifiers present.

    'Storage Order' : str
        Default = 'OBSVAR'

        This option controls how the design matrix, X, should be stored
        in its output array and only has an effect if the design matrix
        is being constructed using ``lm_design_matrix``.

        If 'Storage Order' = 'OBSVAR', X_{ij}, the value for the jth
        variable of the ith observation of the design matrix is stored
        in x[i-1,j-1].

        If 'Storage Order' = 'VAROBS', X_{ij}, the value for the jth
        variable of the ith observation of the design matrix is stored
        in x[j-1,i-1].

        Where x is the output argument of the same name in
        ``lm_design_matrix``.

    'Subject' : str
        This argument gives the subject terms associated with the
        `formula` in a linear mixed effects model.

        The supplied value must consist of a single term, representing
        either a single independent variable, or a single interaction
        term between two or more independent variables.
        All variables in the subject term must not also appear in the
        model formula.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `hform` is not a null Handle or a recognised G22
            handle.

        (`errno` 21)
            The formula contained a mismatched parenthesis.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 22)
            An operator was missing.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 23)
            Invalid use of an operator.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 24)
            Invalid specification for the power operator.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 25)
            Invalid specification for the colon operator.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 26)
            Invalid specification for the mean.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 27)
            Invalid variable name.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 28)
            Missing variable name.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 29)
            After processing, the model contains no terms.

        (`errno` 30)
            An invalid contrast specifier has been supplied.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 41)
            On entry, an invalid option was supplied in `formula`.

        (`errno` 42)
            On entry, an option was supplied in `formula`, but the
            expected delimiter '=' was not found.

        (`errno` 43)
            On entry, an option was supplied in `formula`, but the
            supplied optval was invalid.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            A term contained a repeated variable with a different
            contrast specifier.

    Notes
    -----

    Background
    ~~~~~~~~~~
    Let D denote a data matrix with n observations on m_d independent
    variables, denoted V_1,V_2,...,V_m_d.
    Let y denote a vector of n observations on a dependent variable.

    A linear model, M, as the term is used in this function, expresses a
    relationship between the independent variables, V_j, and the
    dependent variable.
    This relationship can be expressed as a series of additive terms
    T_1+T_2+..., with each term, T_t, representing either a single
    independent variable V_j, called the main effect of V_j, or the
    interaction between two or more independent variables.
    An interaction term, denoted here using the . operator, allows the
    effect of an independent variable on the dependent variable to
    depend on the value of one or more other independent variables.
    As an example, the three-way interaction between V_1,V_2 and V_3 is
    denoted V_1.V_2.V_3 and describes a situation where the effect of
    one of these three variables is influenced by the value of the other
    two.

    This function takes a description of M, supplied as a text string
    containing a formula, and outputs a G22 handle to an internal data
    structure.
    This G22 handle can then be passed to ``lm_design_matrix`` to
    produce a design matrix for use in analysis functions from other
    modules, for example the regression functions of submodule
    ``correg``.

    A more detailed description of what is meant by a G22 handle can be
    found in [the G22 Introduction].

    Syntax
    ~~~~~~
    In its most verbose form M can be described by one or more variable
    names, V_j, and the two operators, + and ..
    In order to allow a wide variety of models to be specified compactly
    this syntax is extended to six operators (+, ., *, -, :, ^) and
    parentheses.

    A formula describing the model is supplied to ``lm_formula`` via a
    character string which must obey the following rules:

    (1) Variables can be denoted by arbitrary names, as long as

        (i) The names used are a subset of those supplied to
            ``lm_describe_data`` when describing D.

        (#) The names do not contain any of the characters in +.*-:^()@.

    (#) The . operator denotes an interaction between two or more
        variables or terms, with V_1.V_2.V_3 denoting the three-way
        interaction between V_1, V_2 and V_3.

    (#) A term in M can contain one or more variable names, separated
        using the . operator, i.e., a term can be either a main effect
        or an interaction term between two or more variables.

        (i) If a variable appears in an interaction term more than once,
            all subsequent appearances, after the first, are ignored,
            therefore, V_1.V_2.V_1 is the same as V_1.V_2.

        (#) The ordering of the variables in an interaction term is
            ignored when comparing terms, therefore, V_1.V_2 is the same
            as V_2.V_1. This ordering may have an effect when the
            resulting G22 handle is passed to another function, for
            example ``lm_design_matrix``.

        (#) Applying the . operator to two terms appends one to the
            other, for example, if T_1 = V_1.V_2 and T_2 = V_3.V_4,
            T_1.T_2 = V_1.V_2.V_3.V_4.

    (#) The + operator allows additional terms to be included in M,
        therefore, T_1+T_2 is a model that includes terms T_1 and T_2.

        (i) If a term is added to M more than once, all subsequent
            appearances, after the first, are ignored, therefore,
            T_1+T_2+T_1 is the same as T_1+T_2.

        (#) The ordering of the terms is ignored whilst parsing the
            formula, therefore, T_1+T_2 is the same as T_2+T_1. This
            ordering may have an effect when the resulting G22 handle is
            passed to another function, for example
            ``lm_design_matrix``.

        (#) Internally, the terms are reordered so that all main effects
            come first, followed by two-way interactions, then three-way
            interactions, etc. The ordering within each of these
            categories is preserved.

    (#) The * operator can be used as a shorthand notation denoting the
        main effects and all interactions between the variables
        involved. Therefore, T_1*T_2 is equivalent to T_1+T_2+T_1.T_2
        and T_1*T_2*T_3 is equivalent to
        T_1+T_2+T_3+T_1.T_2+T_1.T_3+T_2.T_3+T_1.T_2.T_3.

    (#) The - operator removes a term from M, therefore,
        T_1*T_2*T_3-T_1.T_2.T_3 is equivalent to
        T_1+T_2+T_3+T_1.T_2+T_1.T_3+T_2.T_3 as the three-way
        interaction, T_1.T_2.T_3, usually present due to T_1*T_2*T_3 has
        been removed.

    (#) The : operator is a shorthand way of specifying a series of
        variables, with V_1:V_j being equivalent to V_1+V_2+...+V_j.

        (i) This operator can only be used if the variable names end in
            a numeric, therefore, VAR2:VAR4 would be valid, but
            FVAR:LVAR would not.

        (#) The root part of both variable names (i.e., the part before
            the trailing numeric, so VAR in the valid example above)
            must be the same.

        (#) The trailing numeric parts of the two variable names must be
            in ascending order.

    (#) The ^ operator is a shorthand notation for a series of *
        operators. (T_1+T_2+T_3)^2 is equivalent to
        (T_1+T_2+T_3)*(T_1+T_2+T_3) which in turn is equivalent to
        T_1+T_2+T_3+T_1.T_2+T_1.T_3+T_2.T_3.

        (i) This notation is present primarily for use with the :
            operator in examples of the form, (V_1:V_5)^3 which
            specifies a model containing the main effects for variables
            V_1 to V_5 as well as all two - and three-way interactions.

        (#) Using the ^ operator on a single term has no effect,
            therefore, T_2^2 is the same as T_2.

    Precedence
    ++++++++++
    Each operator has an associated default precedence, but this can be
    overridden through the use of parentheses.
    The default precedence is:

    (1) The : operator, with the resulting expression is treated as if
        it was surrounded by parentheses. Therefore, V_1+V_3:V_6*V_7 is
        equivalent to V_1+(V_3+V_4+V_5+V_6)*V_7.

    (#) The ^ operator, with the resulting expression is treated as if
        it was surrounded by parentheses. Therefore, (T_1+T_2+T_3)^2.T_4
        is equivalent to ((T_1+T_2+T_3)^2).T_4, which is the equivalent
        to T_1.T_4+T_2.T_4+T_3.T_4+T_1.T_2.T_4+T_1.T_3.T_4+T_2.T_3.T_4.

    (#) The . operator, so T_1*T_2.T_3 is equivalent to T_1*(T_2.T_3).

    (#) The * operator.

        (i) When using parentheses with the * or . operators the usual
            rules of multiplication apply, therefore,
            (T_1+T_3.T_4).(T_5+T_7) is equivalent to
            T_1.T_5+T_1.T_7+T_3.T_4.T_5+T_3.T_4.T_7 and
            (T_1+T_3.T_4)*(T_5+T_7) is equivalent to
            T_1+T_5+T_7+T_3.T_4+T_1.T_5+T_1.T_7+T_3.T_4.T_5+T_3.T_4.T_7.

        (#) Syntax of the following form is invalid: T_1 o(T_2)oT_3,
            where o indicates an operator, unless one or more of those
            operators are + and/or -. Therefore, T_1.(T_2+T_3)*T_4 is
            invalid, whilst T_1.(T_2+T_3)+T_4 is valid.

    (#) The + and - operators have equal precedence.

        (i) If the terms associated with a - operator do not occur in
            the current expression they are ignored, therefore,
            T_1+(T_2-T_1) is the equivalent to T_1+T_2; the (T_2-T_1)
            part of the expression is calculated first and results in
            T_2 as the T_1 term does not exist in this particular
            sub-expression so cannot be removed.

    Mean Effect / Intercept Term
    ++++++++++++++++++++++++++++
    A mean effect (or intercept term) can be explicitly added to a
    formula by specifying 1 and can be explicitly excluded from the
    formula by specifying -1.
    For example, 1+V_1+V_2 indicates a model with the main effects of
    two variables and a mean effect, whereas V_1+V_2-1 denotes the same
    model, but without the mean effect.
    The mean indicator can appear anywhere in the formula string as long
    as it is not contained within parentheses.

    If the mean effect is not explicitly mentioned in the model formula,
    the model is assumed to include a mean effect.

    Optional Parameters
    ~~~~~~~~~~~~~~~~~~~
    ``lm_formula`` accepts a number of optional parameters described in
    [Other Parameters].
    Usually these parameters are set via call to ``optset``, however
    when specifying a subject term in a mixed effects linear regression
    model it is often more convenient to supply the information along
    with the rest of the formula.
    Therefore, writeable optional parameters can be set via the
    `formula` argument.
    The delimiter / must be used between the main formula and the
    optional parameter.
    For example, supplying a formula of the form V_1+V_2/SUBJECT =
    V_3.V_4, would specify a model formula of V_1+V_2 and set the
    optional parameter 'Subject' to V_3.V_4.

    See Also
    --------
    :meth:`naginterfaces.library.examples.blgm.lm_formula_ex.main`
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.blgm.lm_formula'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], None, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g22yaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_HANDLE, # Input/Output, hform.
        utils._BYREF_STRING, # Input, formula(len=*).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of formula.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _hform = utils._EngineCptrScalarType(
        hform, {'fun_name': fun_name, 'entity': 'hform'},
    ).to_ctypes()
    _formula = utils._EngineCharScalarType(
        formula, {'fun_name': fun_name, 'entity': 'formula'}, exp_chrlen=None,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _hform,
        _formula,
        _errbuf,
        _ierr,
        _formula._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, hform is not a null Handle or a recognised G22 handle.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The formula contained a mismatched parenthesis.',
                    'The position in the formula string of the error is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An operator was missing.',
                    'The position in the formula string of the error is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 23 and _ifmt == 23:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid use of an operator.',
                    'The position in the formula string of the error is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 24 and _ifmt == 24:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid specification for the power operator.',
                    'The position in the formula string of the error is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 25 and _ifmt == 25:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid specification for the colon operator.',
                    'The position in the formula string of the error is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 26 and _ifmt == 26:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid specification for the mean.',
                    'The position in the formula string of the error is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 27 and _ifmt == 27:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid variable name.',
                    'The position in the formula string of the error is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 28 and _ifmt == 28:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Missing variable name.',
                    'The position in the formula string of the error is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 29 and _ifmt == 29:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After processing, the model contains no terms.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 30 and _ifmt == 30:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An invalid contrast specifier has been supplied.',
                    'The position in the formula string of the error is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A term contained a repeated variable with a',
                    'different contrast specifier.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, an invalid option was supplied in formula.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, an option was supplied in formula,',
                    'but the expected delimiter \'=\' was not found.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 43 and _ifmt == 43:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, an option was supplied in formula,',
                    'but the supplied optval was invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def lm_describe_data(hddesc, nobs, nvar, levels, lvnames, vnames):
    u"""
    Describe a dataset.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``, ``optget``.

    ``lm_describe_data`` describes a data matrix.

    For full information please refer to the NAG Library document for
    g22yb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/g22/g22ybf.html

    Parameters
    ----------
    hddesc : Handle, modified in place
        `On entry`: must be set to a null Handle, alternatively an
        existing G22 handle may be supplied in which case this function
        will destroy the supplied G22 handle as if ``handle_free`` had
        been called.

        `On exit`: holds a G22 handle to the internal data structure
        containing a description of the data matrix, D. You **must not**
        change the G22 handle other than through the functions in
        submodule ``blgm``.

    nobs : int
        n, the number of observations in the data matrix, D.

    nvar : int
        m_d, the number of variables in the data matrix, D.

    levels : int, shape (`nvar`)
        `levels`[j-1] contains the number of levels associated with the
        jth variable of the data matrix, for j = 1,2,...,`nvar`.

        If the jth variable is binary, ordinal or continuous,
        `levels`[j-1] should be set to 1; otherwise `levels`[j-1] should
        be set to the number of levels associated with the jth variable
        and the corresponding column of the data matrix is assumed to
        take the value 1 to `levels`[j-1].

    lvnames : int
        The number of variable names supplied in `vnames`.

    vnames : None or str, shape (`lvnames`)
        If `vnames` is not None, `vnames`[j-1] must contain the name of
        the jth variable, for j = 1,2,...,`nvar`. If `lvnames` = 0,
        `vnames` is not referenced.

        The names supplied in `vnames` should be at most 50 characters
        long and be unique.

        If a name longer than 50 characters is supplied it will be
        truncated.

        Variable names must not contain any of the characters +.*-:^()@.

    Other Parameters
    ----------------
    'Number of Observations' : int
        n, the number of observations in the data matrix.

    'Number of Variables' : int
        If queried, this option will return m_d, the number of variables
        in the data matrix.

    'Storage Order' : str
        Default = 'OBSVAR'

        This option states how the data matrix, D, will be stored in its
        input array.

        If 'Storage Order' = 'OBSVAR', D_{ij}, the value for the jth
        variable of the ith observation of the data matrix is stored in
        dat[i-1,j-1].

        If 'Storage Order' = 'VAROBS', D_{ij}, the value for the jth
        variable of the ith observation of the data matrix is stored in
        dat[j-1,i-1].

        Where dat is the input argument of the same name in
        ``lm_design_matrix``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `hddesc` is not a null Handle or a recognised G22
            handle.

        (`errno` 21)
            On entry, `nobs` = *<value>*.

            Constraint: `nobs` >= 0.

        (`errno` 31)
            On entry, `nvar` = *<value>*.

            Constraint: `nvar` >= 0.

        (`errno` 41)
            On entry, j = *<value>* and `levels`[j-1] = *<value>*.

            Constraint: `levels`[i-1] >= 1.

        (`errno` 51)
            On entry, `lvnames` = *<value>* and `nvar` = *<value>*.

            Constraint: `lvnames` = 0 or `nvar`.

        (`errno` 61)
            On entry, variable name i contains one more invalid
            characters, i = *<value>*.

        (`errno` 62)
            On entry, variable names i and j are not unique, i =
            *<value>* and j = *<value>*.

        (`errno` 63)
            On entry, variable names i and j are not unique (possibly
            due to truncation), i = *<value>* and j = *<value>*.

            Maximum variable name length is 50.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 64)
            At least one variable name was truncated to 50 characters.
            Each truncated name is unique and will be used in all
            output.

    Notes
    -----
    Let D denote a data matrix with n observations on m_d independent
    variables, denoted V_1,V_2,...,V_m_d.
    The jth independent variable, V_j can be classified as either
    binary, categorical, ordinal or continuous, where:

    Binary
        V_j can take the value 1 or 0.

    Categorical
        V_j can take one of L_j distinct values or levels. Each level
        represents a discrete category but does not necessarily imply an
        ordering. The value used to represent each level is, therefore,
        arbitrary and, by convention and for convenience, is taken to be
        the integers from 1 to L_j.

    Ordinal
        As with a categorical variable V_j can take one of L_j distinct
        values or levels. However, unlike a categorical variable, the
        levels of an ordinal variable imply an ordering and hence the
        value used to represent each level is not arbitrary. For
        example, V_j = 4 implies a value that is twice as large as V_j =
        2.

    Continuous
        V_j can take any real value.

    ``lm_describe_data`` returns a G22 handle containing a description
    of a data matrix, D.
    The data matrix makes no distinction between binary, ordinal or
    continuous variables.

    A name can also be assigned to each variable.
    If names are not supplied then the default vector of names,
    {'V1','V2',...} is used.

    See Also
    --------
    :meth:`naginterfaces.library.examples.blgm.lm_formula_ex.main`
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.blgm.lm_describe_data'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], None, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g22ybft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_HANDLE, # Input/Output, hddesc.
        utils._BYREF_INT, # Input, nobs.
        utils._BYREF_INT, # Input, nvar.
        utils._EngineIntArrayType.c_type, # Input, levels[nvar].
        utils._BYREF_INT, # Input, lvnames.
        utils._EngineCharArrayType.c_type, # Input, vnames(len=*)[lvnames].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of vnames.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _hddesc = utils._EngineCptrScalarType(
        hddesc, {'fun_name': fun_name, 'entity': 'hddesc'},
    ).to_ctypes()
    _nobs = utils._EngineIntScalarType(
        nobs, {'fun_name': fun_name, 'entity': 'nobs'},
    ).to_ctypes()
    _nvar = utils._EngineIntScalarType(
        nvar, {'fun_name': fun_name, 'entity': 'nvar'},
    ).to_ctypes()
    _nvar_py = utils._EngineIntScalarType.to_py(_nvar)
    _levels = utils._EngineIntArrayType(
        levels, {'fun_name': fun_name, 'entity': 'levels'},
        exp_shape=(_nvar_py,),
    ).to_ctypes()
    _lvnames = utils._EngineIntScalarType(
        lvnames, {'fun_name': fun_name, 'entity': 'lvnames'},
    ).to_ctypes()
    _lvnames_py = utils._EngineIntScalarType.to_py(_lvnames)
    if _lvnames_py != 0:
        if vnames is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'vnames'},
                'must not be None when lvnames != 0',
            )
    _vnames_cls = utils._EngineCharArrayType(
        vnames, {'fun_name': fun_name, 'entity': 'vnames'}, exp_chrlen=None,
        exp_shape=(_lvnames_py,),
        check_contents=_lvnames_py != 0,
        allow_none=True,
    )
    _vnames = _vnames_cls.to_ctypes(can_reference_contents=_lvnames_py != 0)
    _vnames_chrlen = _vnames_cls.chrlen
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _hddesc,
        _nobs,
        _nvar,
        _levels,
        _lvnames,
        _vnames,
        _errbuf,
        _ierr,
        _vnames_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, hddesc is not a null Handle or a recognised G22 handle.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nobs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nobs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nvar >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, j = {:s} and levels[j-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: levels[i-1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lvnames = {:s} and nvar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lvnames = 0 or nvar.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, variable name i contains one more invalid characters,',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 62 and _ifmt == 62:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, variable names i and j are not unique,',
                    'i = {:s} and j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 63 and _ifmt == 63:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, variable names i and j are not unique',
                    '(possibly due to truncation), i = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'and j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Maximum variable name length is 50.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 64 and _ifmt == 64:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'At least one variable name was truncated to 50 characters.',
                    'Each truncated name is unique and will be used in all output.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def lm_design_matrix(hform, hddesc, dat, hxdesc, x, xsize):
    u"""
    Construct a design matrix from a linear model specified using
    ``lm_formula``.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``, ``optget``.

    ``lm_design_matrix`` generates a design matrix from a data matrix
    and model description.

    For full information please refer to the NAG Library document for
    g22yc

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/g22/g22ycf.html

    Parameters
    ----------
    hform : Handle
        A G22 handle to the internal data structure containing a
        description of the model M as returned in hform by
        ``lm_formula``.

    hddesc : Handle
        A G22 handle to the internal data structure containing a
        description of the data matrix, D as returned in hddesc by
        ``lm_describe_data``.

    dat : float, shape (:, :)
        The data matrix, D. By default D_{ij}, the ith value for the jth
        variable, for j = 1,2,...,m_d, for i = 1,2,...,n, should be
        supplied in `dat`[i-1,j-1].

        If the option 'Storage Order', described in
        ``lm_describe_data``, is set to 'VAROBS', D_{ij} should be
        supplied in `dat`[j-1,i-1].

    hxdesc : Handle, modified in place
        `On entry`: must be set to a null Handle, alternatively an
        existing G22 handle may be supplied in which case this function
        will destroy the supplied G22 handle as if ``handle_free`` had
        been called.

        `On exit`: holds a G22 handle to the internal data structure
        containing a description of the design matrix, X. You **must
        not** change the G22 handle other than through the functions in
        submodule ``blgm``.

    x : None or float, shape (:, :), modified in place
        `On exit`, if not None on entry: the design matrix, X. By
        default X_{ij}, the ith value for the jth column, for j =
        1,2,...,m_x, for i = 1,2,...,n, is returned in `x`[i-1,j-1].

        If the option 'Storage Order', described in ``lm_formula``, is
        set to 'VAROBS', X_{ij} is returned in `x`[j-1,i-1].

        If `x` is too small to hold X or `x` is None, the number of
        columns required to hold the design matrix is returned in `mx`.
        The actual shape needed by `x` is returned in `xsize`.

        Under some conditions it is possible to use the data matrix in
        place of the design matrix.

        Specifically, if D has no categorical variables, M has only main
        effects and either has no mean effect or the mean effect does
        not need to be explicitly added to the design matrix.

        If `x` is too small under such circumstances, `errno` = 71 is
        returned and `hxdesc` is set up in such a way as to allow `dat`
        to be used as the design matrix.

    xsize : int, shape (2), modified in place
        `On exit`: the actual shape needed by `x`.

    Returns
    -------
    mx : int
        The minimum number of columns required to hold the design
        matrix.

        In most cases `mx` = m_x.

        The one exception is when `errno` = 71, that is the size of `x`
        was too small but the data matrix given in `dat` can be used as
        the design matrix.

        In this case `mx` holds the number of columns that would be
        required if only the relevant parts of `dat` were copied into a
        new array.

    Other Parameters
    ----------------
    'Formula' : str
        This option returns a verbose formula string describing the
        model, M, used to create the design matrix.
        This formula will only contain variable names, the operators '+'
        and '.' and any contrast identifiers present.

    'Min Number of Columns' : int
        This option returns the minimum number of columns required to
        hold the design matrix, X.
        In most cases 'Min Number of Columns' = 'Number of Columns'.
        The one exception is when `errno` = 71, that is the size of `x`
        was too small but the data matrix given in `dat` can be used as
        the design matrix.
        In this case, 'Number of Columns' = m_x = m_d and 'Min Number of
        Columns' holds the number of columns that would be required if
        only the relevant parts of `dat` were copied into a new array.

    'Number of Columns' : int
        This option returns m_x, the number of columns in the design
        matrix.

    'Number of Observations' : int
        This option returns n, the number of observations in the design
        matrix.

    'Storage Order' : str
        This option returns how the design matrix, X, is stored in `x`.

        If 'Storage Order' = 'OBSVAR', X_{ij}, the value for the jth
        variable of the ith observation of the design matrix is stored
        in `x`[i-1,j-1].

        If 'Storage Order' = 'VAROBS', X_{ij}, the value for the jth
        variable of the ith observation of the design matrix is stored
        in `x`[j-1,i-1].

        It should be noted that 'Storage Order' is not writeable.
        If you wish to change the storage order of the design matrix you
        need to change 'Storage Order' in `hform` as described in [Other
        Parameters for lm_formula] prior to calling
        ``lm_design_matrix``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            `hform` has not been initialized or is corrupt.

        (`errno` 12)
            `hform` is not a G22 handle as generated by ``lm_formula``.

        (`errno` 13)
            A variable name used when creating `hform` is not present in
            `hddesc`.

            Variable name: *<value>*.

        (`errno` 21)
            `hddesc` has not been initialized or is corrupt.

        (`errno` 22)
            `hddesc` is not a G22 handle as generated by
            ``lm_describe_data``.

        (`errno` 31)
            On entry, column j of the data matrix, D, is not consistent
            with information supplied in `hddesc`, j = *<value>*.

        (`errno` 41)
            On entry, n = *<value>* and lddat = *<value>*.

            Constraint: lddat >= n.

        (`errno` 42)
            On entry, m_d = *<value>* and lddat = *<value>*.

            Constraint: lddat >= m_d.

        (`errno` 51)
            On entry, m_d = *<value>* and sddat = *<value>*.

            Constraint: sddat >= m_d.

        (`errno` 52)
            On entry, n = *<value>* and sddat = *<value>*.

            Constraint: sddat >= n.

        (`errno` 61)
            On entry, `hxdesc` is not a null Handle or a recognised G22
            handle.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 14)
            The model contains categorical variables, but no intercept
            or main effects terms have been requested.

            Please check the design matrix returned matches the model
            you require.

        (`errno` 32)
            Column j of the data matrix, D, required rounding more than
            expected when being treated as a categorical variable, j =
            *<value>*.

    Notes
    -----
    ``lm_design_matrix`` generates a design matrix from a data matrix
    and a model description.
    Design matrices encapsulate the observed values of the independent
    variables and the required model in a form that can be used by many
    of the model fitting functions available in the NAG Library, for
    example those in submodule ``correg``.

    Notation
    ~~~~~~~~
    Let D denote a data matrix with n observations on m_d independent
    variables, denoted by V_j, for j = 1,2,...,m_d.
    If V_j is a categorical variable, let L_j denote the number of
    levels associated with it.
    If V_j is a binary, ordinal or continuous variable, let L_j = 1.

    Let V_{ji} denote the ith value of V_j.

    Let M denote a model made up of one or more terms, denoted by T_i.
    Each term consists of either a main effect or an interaction and
    hence can be described using one or more variable names V_j and the
    interaction operator '.'.
    The operator '+' is used to denote the addition of a term to the
    model.
    Therefore, M = T_1+T_2+T_3 = V_1+V_2+V_1.V_2 denotes a model with
    three terms, the first two terms being the main effects for
    variables V_1 and V_2 and the last term the interaction between
    them.
    For simplicity we reorder the terms of the model by the number of
    variables in them, so main effects come first, then two-way
    interactions, then three-way interactions etc.
    By default it is assumed that the model M contains a mean effect (or
    intercept term), if the mean effect is excluded, this will be
    denoted by '-1', so M = T_1 is a model with one term and a mean
    effect and M = T_1-1 is the same model with the mean effect dropped.

    ``lm_design_matrix`` generates an n*m_x design matrix, X, from D and
    M.

    Dummy Variables
    ~~~~~~~~~~~~~~~
    When constructing a design matrix, we cannot work directly with
    categorical variables.
    Categorical variables must first be recoded into dummy variables.
    A categorical variable V_j requires L_j dummy variables.
    Let D^j denote an n*L_j matrix of dummy variables for V_j defined as

        D_li^j = { 1 ; if V_{ji} = l, ; 0 ; otherwise

    where D_l^j is the lth column of D^j and D_li^j is the ith element
    of D_l^j.

    For a binary, ordinal or continuous variable, D_1i^j = V_{ji}.

    Full Design Matrix
    ~~~~~~~~~~~~~~~~~~
    Given a model, M, and the matrices of dummy variables constructing
    the full design matrix X_F is trivial.
    Each term is processed in order and

    (1) If term i is a main effect, that is T_i = V_j for some j, D^j is
        copied into X_F.

    (#) If term i is a two-way interaction, that is T_i = V_j.V_k, for
        some j != k, then

        (i) Loop over l_j = 1,2,...L_j.

        (#) Loop over l_k = 1,2,...L_k.

        (#) Add a column to X_F corresponding to the element-wise
            product of D_l_j^j and D_l_k^k.

    (#) Higher interaction terms are handled in a similar manner as the
        two-way interactions by adding columns constructed from
        multiplying all combinations of the columns of the corresponding
        Ds that correspond to the variables involved. In all cases, the
        variables towards the right hand side of a term are iterated
        over the quickest.

    Contrasts
    ~~~~~~~~~
    Using the full design matrix X_F in an analysis can result in an
    overparameterized model.
    This is due to X_F often not being of full rank as the sum of all
    the dummy variables for a particular variable is a vector of ones.
    This source of overparameterization can be alleviated by using a
    design matrix X where (some) dummy variables are replaced by
    contrasts.
    For a categorical variable V_j the contrasts are a set of L_j-1
    functionally independent linear combinations of the dummy variables.

    Whilst the choice of contrasts used in term T_i will affect the
    individual model coefficients (parameters), it has no effect on the
    overall contribution of T_i.

    For a given variable V_j, the contrasts can be represented by an
    L_j*L_j-1 matrix, C_j.
    The rows of C_j correspond to a particular value of V_j and the
    columns correspond to the values to use in the design matrix.

    Six types of contrast are available in ``lm_design_matrix``; two
    types of treatment contrasts, two types of sum contrasts, Helmert
    contrasts and polynomial contrasts.
    Unless specified otherwise, the contrasts used by
    ``lm_design_matrix`` are treatment contrasts relative to the first
    level.
    See the description of the option 'Contrast' in ``lm_formula`` for
    ways of changing the contrasts used.

    Treatment Contrasts
    +++++++++++++++++++
    Treatment contrasts are taken relative to either the first or last
    level of the variable.
    For example, if L_j = 4,

        C_j = [table omitted]

    would be the contrast matrix for V_j using treatment contrasts
    relative to the first level.
    The contrast matrix obtained when using treatment contrasts relative
    to the last level is similar, but the row of zeros appears at the
    bottom and all other rows are shifted up one.

    Strictly speaking, the term `contrast` implies that each row in the
    contrast matrix sums to zero.
    That is not the case for treatment contrasts, however they are
    included as this coding is commonly used in practice.

    Sum Contrasts
    +++++++++++++
    Sum contrasts are similar to treatment contrasts and again can be
    taken relative to the first or last level of the variable.
    Unlike treatment contrasts, sum contrasts effectively constrain the
    coefficients related to the variable to sum to zero.
    For example, if L_j = 4,

        C_j = [table omitted]

    would be the contrast matrix for V_j using treatment contrasts
    relative to the last level.
    The contrast matrix obtained when using treatment contrasts relative
    to the first level is similar, but the row of -1s appears at the top
    and all other rows are shifted down one.

    Helmert Contrasts
    +++++++++++++++++
    With Helmert contrasts level l of the variable is compared with the
    average effect of all previous levels.
    For example, if L_j = 4,

        C_j = [table omitted]

    would be the contrast matrix for V_j using Helmert contrasts.

    Polynomial Contrasts
    ++++++++++++++++++++
    With polynomial contrasts the entries in the columns of C_j
    correspond in linear, quadratic, cubic, quartic, etc. terms to a
    hypothetical underlying numeric variable that takes equally spaced
    values at each level.
    For example, if L_j = 4,

        C_j = [table omitted]

    would be the contrast matrix for V_j using polynomial contrasts.

    When Contrasts Can Be Used
    ++++++++++++++++++++++++++
    Depending on the specifics of the model, M, it may not be possible
    to always replace the L_j dummy variables with L_j-1 contrasts for
    all variables in all terms and retain the same model.
    A simple example of this is a data matrix, D, with four observations
    and two variables which have two and three levels respectively.
    This data matrix might look something like:

        D = [table omitted]

    For the sake of argument, assume that our model contains the main
    effect for each variable, but does not contain a mean effect (or
    intercept term).
    So using the notation established earlier, M = V_1+V_2-1.
    The full design matrix, X_F, for this data matrix and model would be

        X_F = [table omitted]

    However, X_F is not of full rank (and hence M is overparameterized)
    because the sum of the first two columns is a vector of ones as is
    the sum of the last three columns.

    In order to alleviate this we might try constructing X_C where the
    dummy variables have been replaced by contrasts.
    Assuming treatment contrasts, relative to the first level, we would
    have

        X_C = [table omitted]

    However, using X_C makes an implicit assumption that the expected
    value of the dependent variable (the quantity being modelled) is
    zero when V_1 = 1 and V_2 = 1.
    This assumption was not made when we used X_F and hence the two
    design matrices are not equivalent.
    One solution would be to use dummy variables for V_1 and contrasts
    for V_2, which would result in a design matrix, X of

        X = [table omitted]

    Using X would give an equivalent model to using X_F.

    The algorithm used by ``lm_design_matrix`` to decide which
    variables, in which terms, can be coded as contrasts and which need
    to be coded as dummy variables is described below.

    Suppose V_j is any variable that appears in term T_i, let T_{i(j)}
    denote the term obtained by dropping V_j from T_i.
    For example, if T_3 = V_1.V_2.V_3, T_{3(2)} = V_1.V_3.
    In this context, the empty term is taken to be the mean effect (or
    intercept term).
    We say that T_{i(j)} appears in M if there exists a term T_k, k < i,
    that contains all of the variables appearing in T_{i(j)}.
    In most cases T_k = T_{i(j)}, but this is not required.
    Note, as stated earlier, the terms in M are ordered by the number of
    variables in them.

    A variable, V_j in term T_i is coded by contrasts if T_{i(j)}
    appears in M and by dummy variables otherwise.
    It is, therefore, possible for variable V_j to be coded by contrasts
    in some terms and dummy variables in others within the same X.

    The above rule assumes the presence of a mean effect.
    If no such effect is present in the model, the main effect of the
    first categorical variable is coded by dummy variables to
    compensate.
    If no main effects appear in the model, the warning `errno` = 14 is
    returned.

    A longer description and informal proof that the resulting X is a
    suitable design matrix for the model of interest can be found in
    module two of Chambers and Hastie (1992).

    Mean Effect
    ~~~~~~~~~~~
    The mean effect (or intercept term) is included in a design matrix
    by adding a column of ones as the first column of X.
    However, many model fitting functions in the NAG Library handle the
    mean effect as a special case and do not require it to be explicitly
    added to the design matrix.
    Therefore, by default, ``lm_design_matrix`` does not explicitly add
    the mean effect to the design matrix.
    This behaviour can be changed via the option 'Explicit Mean' in
    ``lm_formula``.

    References
    ----------
    Chambers, J M and Hastie, T J, 1992, `Statistical Models in S`,
    Wadsworth and Brooks/Cole Computer Science Series

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.blgm.lm_design_matrix'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], None, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g22ycft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_HANDLE, # Input, hform.
        utils._BYREF_HANDLE, # Input, hddesc.
        utils._EngineFloat64ArrayType.c_type, # Input, dat[lddat, sddat].
        utils._BYREF_INT, # Input, lddat.
        utils._BYREF_INT, # Input, sddat.
        utils._BYREF_HANDLE, # Input/Output, hxdesc.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, sdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, sdx.
        utils._BYREF_INT, # Output, mx.
        utils._EngineIntArrayType.c_type, # Output, xsize[2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((dat, 'dat'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _hform = utils._EngineCptrScalarType(
        hform, {'fun_name': fun_name, 'entity': 'hform'},
    ).to_ctypes()
    _hddesc = utils._EngineCptrScalarType(
        hddesc, {'fun_name': fun_name, 'entity': 'hddesc'},
    ).to_ctypes()
    _lddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lddat = utils._EngineIntScalarType(
        _lddat_py, {'fun_name': fun_name, 'entity': '_lddat'},
    ).to_ctypes()
    _sddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sddat = utils._EngineIntScalarType(
        _sddat_py, {'fun_name': fun_name, 'entity': '_sddat'},
    ).to_ctypes()
    _dat = utils._EngineFloat64ArrayType(
        dat, {'fun_name': fun_name, 'entity': 'dat'},
        exp_shape=(None, None),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _hxdesc = utils._EngineCptrScalarType(
        hxdesc, {'fun_name': fun_name, 'entity': 'hxdesc'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _sdx = utils._EngineIntScalarType(
        _sdx_py, {'fun_name': fun_name, 'entity': '_sdx'},
    ).to_ctypes()
    if _ldx_py > 0 and _sdx_py > 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when ldx > 0 and sdx > 0',
            )
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(None, None),
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _mx = utils._EngineIntScalarType.empty_ctypes()
    _xsize_cls = utils._EngineIntArrayType(
        xsize, {'fun_name': fun_name, 'entity': 'xsize'},
        exp_shape=(2,),
        check_contents=False,
    )
    _xsize = _xsize_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _hform,
        _hddesc,
        _dat,
        _lddat,
        _sddat,
        _hxdesc,
        _x,
        _ldx,
        _sdx,
        _mx,
        _xsize,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hform has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hform is not a G22 handle as generated by lm_formula.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A variable name used when creating hform is not present in hddesc.',
                    'Variable name: {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The model contains categorical variables, but no intercept',
                    'or main effects terms have been requested.',
                    'Please check the design matrix returned matches the model you require.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hddesc has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hddesc is not a G22 handle as generated by lm_describe_data.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, column j of the data matrix, D, is not',
                    'consistent with information supplied in hddesc,',
                    'j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Column j of the data matrix, D, required rounding',
                    'more than expected when being treated as a categorical variable,',
                    'j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and lddat = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lddat >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m_d = {:s} and lddat = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lddat >= m_d.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m_d = {:s} and sddat = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: sddat >= m_d.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and sddat = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: sddat >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, hxdesc is not a null Handle or a recognised G22 handle.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineNoErrorExit()
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineNoErrorExit()
        elif _ierr_py == 82 and _ifmt == 82:
            _err_class = utils._EngineNoErrorExit()
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineNoErrorExit()
        elif _ierr_py == 92 and _ifmt == 92:
            _err_class = utils._EngineNoErrorExit()
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        if x is not None:
            _x_cls.refresh_py(_x)
        _xsize_cls.refresh_py(_xsize)
        _mx_py = utils._EngineIntScalarType.to_py(_mx)
        return_data = _mx_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['mx'],
            )(
                _mx_py,
            )
    _err_class.manage()
    return return_data

def lm_submodel(what, hform, hxdesc, lisx, isx, lplab, plab, lvinfo, vinfo):
    u"""
    Construct a vector indicating which columns of a design matrix to
    include in a submodel specified using ``lm_formula``.

    ``lm_submodel`` produces labels for the columns of a design matrix,
    model parameters and a vector of column inclusion flags suitable for
    use with functions in submodule ``correg``.
    Thus allowing for submodels to be fit using the same design matrix.

    For full information please refer to the NAG Library document for
    g22yd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/g22/g22ydf.html

    Parameters
    ----------
    what : str
        Controls what labels are to be produced:

        `what` = 'S'

            Labels for a submodel are required. The submodel must be
            supplied in `hform`.

        `what` = 'X'

            Labels for the design matrix X.

            If `hxdesc` was returned by ``correg.lmm_init`` in hlmm then
            X is the design matrix associated with the fixed parameters.

        `what` = 'Z'

            Labels for the design matrix Z.

            If `hxdesc` was returned by ``correg.lmm_init`` in hlmm then
            Z is the part of the design matrix associated with the
            random parameters.

        `what` = 'V'

            Labels for the variance components.

    hform : Handle
        A G22 handle to the internal data structure containing a
        description of the required submodel M_S, as returned in hform
        by ``lm_formula``. If `what` != 'S' `hform` is not referenced
        and need not be set.

    hxdesc : Handle
        A G22 handle to the internal data structure containing a
        description of the design matrix, D.

    lisx : int
        Length of `isx`.

    isx : None or int, shape (`lisx`), modified in place
        `On exit`, if not None on entry: if `lisx` != 0, an array
        indicating which columns of the design matrix from the model
        specified in `hform` are to be used.

        `isx`[j-1] = 0

            The jth column of the design matrix, X, should not be
            included in the analysis.

        `isx`[j-1] = 1

            The jth column of the design matrix, X, should be included
            in the analysis.

        If `lisx` = 0, `isx` is not referenced.

    lplab : int
        The length of `plab`.

        As p <= m_x+1, if labels are required, using `lplab` = m_x+1
        will always be sufficient.

    plab : None or str, shape (`lplab`), modified in place
        `On exit`, if not None on entry: if `lplab` != 0, the names
        associated with the p parameters in the model.

        If `intcpt` = 'Z', the labels in `plab` are also the labels for
        the columns of design matrix used in the analysis.

        If `intcpt` = 'M', columns `plab`[1] to `plab`[p-1] are the
        corresponding column labels.

        If a mean effect is present in M_S, the corresponding label is
        always in `plab`[0].

        If `lplab` = 0, `plab` is not referenced.

    lvinfo : int
        The length of `vinfo`.

        Let n_T denote the number of terms in M_S, n_{Tt} denote the
        number of variables in the tth term and m_{xt} denote the number
        of columns of X corresponding to the tth term.

        The required size of `vinfo`, denoted a is given by:

            a = sum(m_{xt}*(1+3n_{Tt})).

        If the model includes a mean effect, a should be incremented by
        one.

        The values n_T, n_{Tt} and m_{xt} are not trivial to calculate
        as they require the formula describing the model to be fully
        expanded and the contrast / dummy variable encoding to be known.

        Therefore, if `lisx`, `lplab` or `lvinfo` are too small and
        `lvinfo` >= 3, `errno` = 102 is returned and the required sizes
        for these arrays are returned in `vinfo`[0], `vinfo`[1] and
        `vinfo`[2] respectively.

    vinfo : None or int, shape (`lvinfo`), modified in place
        `On exit`, if not None on entry: if `lvinfo` != 0, information
        encoding a description of the parameters in the model.

        The encoding information can be extracted as follows:

        (i) Set k = 1.

        (#) Iterate j from 1 to p.

            (1) Set b = `vinfo`[k-1].

            (#) Increment k.

            (#) Iterate i from 1 to b.

                (a) Set v_i = `vinfo`[k-1].

                (#) Set l_i = `vinfo`[k].

                (#) Set c_i = `vinfo`[k+1].

                (#) Increment k by 3.

            (#) The jth model parameter corresponds to the interaction
                between the b variables held in columns v_1,v_2,...,v_b
                of D. Therefore, b = 1 indicates a main effect, b = 2 a
                two-way interaction, etc..

                If b = 0, the jth model parameter corresponds to the
                mean effect.

                If l_i = 0, the corresponding variable v_i is binary,
                ordinal or continuous.

                Otherwise, l_i is the level for the corresponding
                variable for model parameter j.

                c_i is a numeric flag indicating the contrast used in
                the case of a categorical variable.

                With c_i = 0 indicating that dummy variables were used
                for variable v_i in this term.

                The remaining six types of contrast; treatment contrasts
                (with respect to the first and last levels), sum
                contrasts (with respect to the first and last levels),
                Helmert contrasts and polynomial contrasts, as described
                in ``lm_design_matrix``, are identified by the integers
                one to six respectively.

        If `lvinfo` = 0, `vinfo` is not referenced.

    Returns
    -------
    intcpt : str
        If `intcpt` = 'M', in order to fit the model M_S to D using X,
        any analysis function should include an implicit mean effect
        (intercept term).

        `intcpt` = 'Z', if M_S does not include a mean effect or the
        mean effect has been explicitly included in the design matrix.

    ip : int
        p, the number of parameters in the (sub)model, including the
        intercept if one is present. If `what` = 'S', then the submodel
        is the one specified in `hform` otherwise the model is the one
        used when defining the design matrix described in `hxdesc`.

        If `lisx` != 0, if `intcpt` = 'Z', p = sum_i = 1^m_x`isx`[i-1],
        otherwise p = sum_i = 1^m_x`isx`[i-1]+1.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `what` = *<value>* was an illegal value.

        (`errno` 12)
            Supplied value of `what` is not valid for the G22 handle
            supplied in `hxdesc`.

        (`errno` 21)
            `hform` has not been initialized or is corrupt.

        (`errno` 22)
            `hform` is not a G22 handle as generated by ``lm_formula``.

        (`errno` 23)
            A variable name used when creating `hform` is not present in
            `hxdesc`.

            Variable name: *<value>*.

        (`errno` 24)
            The model and the design matrix are not consistent. The
            design matrix was constructed in the presence of a mean
            effect and the model does not include a mean effect.

        (`errno` 25)
            The model and the design matrix are not consistent. The
            model includes a term not present in the design matrix.

            Term: *<value>*.

        (`errno` 26)
            The model and the design matrix are not consistent.

            Term: *<value>*.

            This is likely due to the design matrix being constructed in
            the presence of either a mean effect or main effect that is
            not present in the model.

        (`errno` 31)
            `hxdesc` has not been initialized or is corrupt.

        (`errno` 32)
            `hxdesc` is not a G22 handle as generated by
            ``lm_design_matrix``.

        (`errno` 61)
            On entry, `lisx` = *<value>* and m_x = *<value>*.

            Constraint: `lisx` = 0 or `lisx` >= m_x.

        (`errno` 81)
            On entry, `lplab` = *<value>* and p = *<value>*.

            Constraint: `lplab` = 0 or `lplab` >= p.

        (`errno` 91)
            On entry, `plab` is too short to hold the parameter labels.
            Long labels will be truncated.

            The longest parameter label is *<value>*.

        (`errno` 101)
            On entry, `lvinfo` is too small.

            `lvinfo` = *<value>*.

            Constraint: `lvinfo` = 0 or `lvinfo` >= *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 27)
            The model and the design matrix are not consistent. The
            model specifies different contrasts to those used when the
            design matrix was constructed. The contrasts specified in
            `hform` will be ignored.

        (`errno` 28)
            The model may not be as expected.

            This is due to the model not containing the categorical
            variable adjusted to account for no mean effect when the
            design matrix was constructed.

        (`errno` 33)
            `hxdesc` has not passed through the model fitting function.

        (`errno` 102)
            On entry, one or more of `lisx`, `lplab` or `lvinfo` are
            nonzero, but too small.

            Minimum values are zero, or *<value>*, *<value>* and
            *<value>* respectively.

            The minimum values are returned in the first three elements
            of `vinfo`.

    Notes
    -----
    ``lm_submodel`` is a utility function for use with ``lm_formula``,
    ``lm_describe_data`` and ``lm_design_matrix``.
    It can be used to construct labels for the columns for an n*m_x
    design matrix, X, created by ``lm_design_matrix`` and return
    additional input vectors and flags required by a number of NAG
    Library model fitting functions.

    Many of the analysis functions that require a design matrix to be
    supplied allow submodels to be defined through the use of a vector
    of ones or zeros indicating whether a column of X should be included
    or excluded from the analyses (see for example isx in
    ``correg.linregm_fit`` or ``correg.glm_normal``).
    This allows nested models to be fit without having to reconstructed
    the design matrix for each analysis.

    Let M denote a model constructed by ``lm_formula``, D a data matrix
    as described by ``lm_describe_data`` and X be the corresponding
    design matrix constructed by ``lm_design_matrix`` from M and D.
    A different model, M_S is a submodel of M if each term in M_S,
    including the mean effect (intercept term) is also present in M.

    If M_S is a submodel of M, you can fit M_S to D using a design
    matrix whose columns are a subset of the columns of X.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.blgm.lm_submodel'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g22ydft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iwhat.
        utils._BYREF_HANDLE, # Input, hform.
        utils._BYREF_HANDLE, # Input, hxdesc.
        utils._BYREF_INT, # Output, iintcpt.
        utils._BYREF_INT, # Output, ip.
        utils._BYREF_INT, # Input, lisx.
        utils._EngineIntArrayType.c_type, # Output, isx[lisx].
        utils._BYREF_INT, # Input, lplab.
        utils._EngineCharArrayType.c_type, # Output, plab(len=*)[lplab].
        utils._BYREF_INT, # Input, lvinfo.
        utils._EngineIntArrayType.c_type, # Output, vinfo[lvinfo].
        utils._BYREF_INT, # Output, mlterm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of plab.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _what = utils._EngineCharScalarType(
        what, {'fun_name': fun_name, 'entity': 'what'}, exp_chrlen=None,
    ).to_ctypes()
    _what_py = utils._EngineCharScalarType.to_py(_what)
    if _what_py.upper() == 'S':
        _iwhat = utils._EngineIntScalarType.c_type(1)
    elif _what_py.upper() == 'X':
        _iwhat = utils._EngineIntScalarType.c_type(2)
    elif _what_py.upper() == 'Z':
        _iwhat = utils._EngineIntScalarType.c_type(3)
    elif _what_py.upper() == 'V':
        _iwhat = utils._EngineIntScalarType.c_type(4)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, what = {:s} was an illegal value.'.format(
                    utils._STR_FMT_STR.format(_what_py),
                ),
            ],
            errcodes=(11,),
        )
        _err_class.manage()
    _hform = utils._EngineCptrScalarType(
        hform, {'fun_name': fun_name, 'entity': 'hform'},
    ).to_ctypes()
    _hxdesc = utils._EngineCptrScalarType(
        hxdesc, {'fun_name': fun_name, 'entity': 'hxdesc'},
    ).to_ctypes()
    _iintcpt = utils._EngineIntScalarType.empty_ctypes()
    _ip = utils._EngineIntScalarType.empty_ctypes()
    _lisx = utils._EngineIntScalarType(
        lisx, {'fun_name': fun_name, 'entity': 'lisx'},
    ).to_ctypes()
    _lisx_py = utils._EngineIntScalarType.to_py(_lisx)
    if _lisx_py != 0:
        if isx is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'isx'},
                'must not be None when lisx != 0',
            )
    _isx_cls = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_lisx_py,),
        check_contents=False,
        allow_none=True,
    )
    _isx = _isx_cls.to_ctypes(can_reference_contents=False)
    _lplab = utils._EngineIntScalarType(
        lplab, {'fun_name': fun_name, 'entity': 'lplab'},
    ).to_ctypes()
    _lplab_py = utils._EngineIntScalarType.to_py(_lplab)
    if _lplab_py != 0:
        if plab is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'plab'},
                'must not be None when lplab != 0',
            )
    _plab_cls = utils._EngineCharArrayType(
        plab, {'fun_name': fun_name, 'entity': 'plab'}, exp_chrlen=None,
        exp_shape=(_lplab_py,),
        check_contents=False,
        allow_none=True,
    )
    _plab = _plab_cls.to_ctypes(can_reference_contents=False)
    _plab_chrlen = _plab_cls.chrlen
    _lvinfo = utils._EngineIntScalarType(
        lvinfo, {'fun_name': fun_name, 'entity': 'lvinfo'},
    ).to_ctypes()
    _lvinfo_py = utils._EngineIntScalarType.to_py(_lvinfo)
    if _lvinfo_py != 0:
        if vinfo is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'vinfo'},
                'must not be None when lvinfo != 0',
            )
    _vinfo_cls = utils._EngineIntArrayType(
        vinfo, {'fun_name': fun_name, 'entity': 'vinfo'},
        exp_shape=(_lvinfo_py,),
        check_contents=False,
        allow_none=True,
    )
    _vinfo = _vinfo_cls.to_ctypes(can_reference_contents=False)
    _mlterm = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iwhat,
        _hform,
        _hxdesc,
        _iintcpt,
        _ip,
        _lisx,
        _isx,
        _lplab,
        _plab,
        _lvinfo,
        _vinfo,
        _mlterm,
        _errbuf,
        _ierr,
        _plab_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Supplied value of what is not valid for the G22 handle supplied in hxdesc.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hform has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hform is not a G22 handle as generated by lm_formula.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 23 and _ifmt == 23:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A variable name used when creating hform is not present in hxdesc.',
                    'Variable name: {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 24 and _ifmt == 24:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The model and the design matrix are not consistent.',
                    'The design matrix was constructed in the presence of a mean effect and',
                    'the model does not include a mean effect.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 25 and _ifmt == 25:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The model and the design matrix are not consistent.',
                    'The model includes a term not present in the design matrix.',
                    'Term: {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 26 and _ifmt == 26:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The model and the design matrix are not consistent.',
                    'Term: {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'This is likely due to the design matrix being constructed in the',
                    'presence of either a mean effect or main effect that is not present',
                    'in the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 27 and _ifmt == 27:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The model and the design matrix are not consistent.',
                    'The model specifies different contrasts to those used when the design matrix',
                    'was constructed. The contrasts specified in hform will be ignored.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 28 and _ifmt == 28:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The model may not be as expected.',
                    'This is due to the model not containing the categorical variable adjusted',
                    'to account for no mean effect when the design matrix was constructed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hxdesc has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hxdesc is not a G22 handle as generated by lm_design_matrix.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 33 and _ifmt == 33:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'hxdesc has not passed through the model fitting function.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lisx = {:s} and m_x = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lisx = 0 or lisx >= m_x.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lplab = {:s} and p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lplab = 0 or lplab >= p.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, plab is too short to hold the parameter labels.',
                    'Long labels will be truncated.',
                    'The longest parameter label is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lvinfo is too small.',
                    'lvinfo = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lvinfo = 0 or lvinfo >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 102 and _ifmt == 102:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, one or more of lisx, lplab or lvinfo are nonzero, but too small.',
                    'Minimum values are zero, or {:s}, {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    '{:s} respectively.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'The minimum values are returned in the first three elements of vinfo.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        if isx is not None:
            _isx_cls.refresh_py(_isx)
        if plab is not None:
            _plab_cls.refresh_py(_plab)
        if vinfo is not None:
            _vinfo_cls.refresh_py(_vinfo)
        _iintcpt_py = utils._EngineIntScalarType.to_py(_iintcpt)
        if _iintcpt_py == 1:
            _intcpt_py = 'M'
        elif _iintcpt_py == 0:
            _intcpt_py = 'Z'
        else:
            raise utils.NagException(
                {'fun_name': fun_name},
                'Unexpected value for internal iintcpt.',
            )
        _ip_py = utils._EngineIntScalarType.to_py(_ip)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['intcpt', 'ip'],
        )(
            _intcpt_py,
            _ip_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def handle_free(handle):
    u"""
    Destroy a G22 handle and deallocate all the memory used.

    ``handle_free`` destroys a G22 handle and deallocates all the memory
    used.

    For full information please refer to the NAG Library document for
    g22za

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/g22/g22zaf.html

    Parameters
    ----------
    handle : Handle, modified in place
        `On entry`: the G22 handle to be destroyed.

        `On exit`: the handle is destroyed and set to a null Handle.

    Raises
    ------
    NagValueError
        (`errno` 12)
            `handle` has been corrupted.

        (`errno` 13)
            `handle` is a handle to an unknown data structure.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 11)
            `handle` has not been initialized.

    Notes
    -----
    Each G22 handle should be deallocated to avoid memory leaks.
    Therefore, ``handle_free`` should be called on all such handles
    which are no longer needed.
    Please note that passing an uninitialized handle might cause
    unpredictable behaviour, including a crash of your program.
    """

    excs = []
    fun_name = 'naginterfaces.base.blgm.handle_free'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], None, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g22zaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_HANDLE, # Input/Output, handle.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _handle = utils._EngineCptrScalarType(
        handle, {'fun_name': fun_name, 'entity': 'handle'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _handle,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'handle has not been initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'handle has been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'handle is a handle to an unknown data structure.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def optset(handle, optstr):
    u"""
    Option setting function for submodule ``blgm``.

    ``optset`` is a general option setting function for functions in
    submodule ``blgm``.
    It can set a single option or reset all of them to their default.

    For full information please refer to the NAG Library document for
    g22zm

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/g22/g22zmf.html

    Parameters
    ----------
    handle : Handle
        The G22 handle which **must** have been initialized by one of
        submodule ``blgm``'s initialization functions.

    optstr : str
        A string identifying the option, its value and, where required,
        the instance identifier.

        Defaults

            Resets all options to their default values.

        option = optval

            Sets (all instances) of option to optval.

        option:instance identifier = optval

            Sets a single instance of option to optval.

        option = default

            Resets (all instances) of option to their default value.

        option:instance identifier = default

            Resets a single instance of option to its default value.

        `optstr` is case insensitive and option, `instance identifier`
        and optval may consist of one or more tokens separated by white
        space.

        See the documentation of the individual functions in [the G22
        Introduction] for details of valid values for option, `instance
        identifier` and optval.

    Raises
    ------
    NagValueError
        (`errno` 11)
            `handle` has not been initialized or is corrupt.

        (`errno` 12)
            `handle` is not a G22 handle.

        (`errno` 21)
            On entry, option was not recognized.

            `optstr` = *<value>*.

        (`errno` 22)
            On entry, the expected delimiter '=' was not found.

            `optstr` = *<value>*.

        (`errno` 23)
            On entry, option is read only.

            `optstr` = *<value>*.

        (`errno` 24)
            On entry, could not convert optval to an integer.

            `optstr` = *<value>*.

        (`errno` 25)
            On entry, could not convert optval to a real.

            `optstr` = *<value>*.

        (`errno` 26)
            On entry, optval is not a valid value for option.

            `optstr` = *<value>*.

        (`errno` 121)
            Invalid `instance identifier` for option.

            On entry, `optstr` = *<value>*.

        (`errno` 122)
            Numeric `instance identifier` is out of range.

            On entry, instance identifier = *<value>*.

            Constraint: *<value>* <= instance identifier and instance
            identifier <= *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 123)
            On entry, option cannot have an associated `instance
            identifier`. The supplied `instance identifier` was ignored.

            `optstr` = *<value>*.

    Notes
    -----
    ``optset`` can only be called on G22 handles.
    Its purpose is to reset all options to their default values or set a
    single option to a user-supplied value.

    Options and their values are, in general, presented as a character
    string of the form 'option = optval'; alphabetic characters can be
    supplied in either upper or lower case. optval will normally be
    either an integer, real or character value as defined in the
    description of the specific option.
    In addition, all options can take an optval DEFAULT which resets the
    option to its default value.

    In cases where an option may have multiple instances an `instance
    identifier` can be specified.
    This is presented using the form 'option:instance identifier =
    optval'.
    In such cases, if the instance identifier is omitted, the value of
    all instances are changed.

    Information relating to available option names, their corresponding
    valid values, whether the use of an instance identifier may be
    appropriate and what form it can take is given in the individual
    function documents.

    See Also
    --------
    :meth:`naginterfaces.library.examples.blgm.lm_formula_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.blgm.optset'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], None, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g22zmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_HANDLE, # Input, handle.
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _handle = utils._EngineCptrScalarType(
        handle, {'fun_name': fun_name, 'entity': 'handle'},
    ).to_ctypes()
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _handle,
        _optstr,
        _errbuf,
        _ierr,
        _optstr._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _optstr_py = utils._EngineCharScalarType.to_py(_optstr)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'handle has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'handle is not a G22 handle.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, option was not recognized.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the expected delimiter \'=\' was not found.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 23 and _ifmt == 23:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, option is read only.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 24 and _ifmt == 24:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, could not convert optval to an integer.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 25 and _ifmt == 25:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, could not convert optval to a real.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 26 and _ifmt == 26:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, optval is not a valid value for option.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 121 and _ifmt == 121:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid instance identifier for option.',
                    'On entry, optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 122 and _ifmt == 122:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Numeric instance identifier is out of range.',
                    'On entry, instance identifier = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: {:s} <= instance identifier'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and instance identifier <= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 123 and _ifmt == 123:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, option cannot have an associated instance identifier. The supplied',
                    'instance identifier was ignored.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def optget(handle, optstr):
    u"""
    Option getting function for submodule ``blgm``.

    ``optget`` is a general option getting function for submodule
    ``blgm``.
    It is used to query the value of options.

    For full information please refer to the NAG Library document for
    g22zn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/g22/g22znf.html

    Parameters
    ----------
    handle : Handle
        The G22 handle which **must** have been initialized by one of
        submodule ``blgm``'s initialization functions.

    optstr : str
        A string identifying the option and, where required, the
        instance identifier.

        **identify**

            Returns a string description of the G22 handle supplied in
            `handle`. See [Further Comments] for more details.

        option

            Returns the value of option. If there are multiple instances
            of option, the value of the first is returned. If not all
            instances of option have the same value, `errno` = 124 is
            returned.

        option:instance identifier

            Returns the value of a single instance of option.

        `optstr` is case insensitive and option and `instance
        identifier` may consist of one or more tokens separated by white
        space.

        See the documentation of the individual submodule ``blgm``
        functions for details of valid values for option and `instance
        identifier`.

    Returns
    -------
    ivalue : int
        If `optype` = 1, option identifies an integer-valued option and
        `ivalue` holds its current value.

    rvalue : float
        If `optype` = 2, option identifies a real-valued option and
        `rvalue` holds its current value.

    cvalue : str
        If `optype` = 3, option identifies a character-valued option and
        `cvalue` holds its current value.

    optype : int
        Indicates whether option is an integer, real or character-valued
        option and hence which of `ivalue`, `rvalue` or `cvalue` holds
        the current value.

        `optype` = 1

            option is an integer-valued option; its current value has
            been returned in `ivalue`.

        `optype` = 2

            option is a real-valued option; its current value has been
            returned in `rvalue`.

        `optype` = 3

            option is a character-valued option; its current value has
            been returned in `cvalue`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            `handle` has not been initialized or is corrupt.

        (`errno` 12)
            `handle` is not a G22 handle.

        (`errno` 21)
            On entry, option was not recognized: `optstr` = *<value>*.

        (`errno` 22)
            On entry, option is not readable: `optstr` = *<value>*.

        (`errno` 121)
            Invalid `instance identifier` for option.

            On entry, `optstr` = *<value>*.

        (`errno` 122)
            Numeric `instance identifier` is out of range.

            On entry, instance identifier = *<value>*.

            Constraint: *<value>* <= instance identifier and instance
            identifier <= *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 123)
            On entry, option cannot have an associated `instance
            identifier`. The supplied `instance identifier` was ignored.

            `optstr` = *<value>*.

        (`errno` 124)
            option has multiple instances. Information from the first
            instance has been returned.

    Notes
    -----
    ``optget`` can only be called on G22 handles.
    It can be used to query the current values of options and will
    return either an integer, real or character value dependent upon the
    type associated with the option being queried.

    The option of interest is presented as a character string of the
    form 'option'

    In cases where an option may have multiple instances in a particular
    G22 handle an `instance identifier` can be specified.
    This is presented using the form 'option:instance identifier'.
    In such cases, if the instance identifier is omitted, the value of
    the first instance is returned.
    If the value of option is not the same for all instances and an
    instance identifier is omitted, a warning is raised.

    Information relating to available option names, their corresponding
    valid values, whether the use of an instance identifier may be
    appropriate and what form it can take is given in the individual
    function documents.
    """

    excs = []
    fun_name = 'naginterfaces.base.blgm.optget'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], None, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g22znft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_HANDLE, # Input, handle.
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._BYREF_INT, # Output, ivalue.
        utils._BYREF_FLOAT64, # Output, rvalue.
        utils._BYREF_STRING, # Output, cvalue(len=*).
        utils._BYREF_INT, # Output, optype.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of cvalue.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _handle = utils._EngineCptrScalarType(
        handle, {'fun_name': fun_name, 'entity': 'handle'},
    ).to_ctypes()
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _ivalue = utils._EngineIntScalarType.empty_ctypes()
    _rvalue = utils._EngineFloat64ScalarType.empty_ctypes()
    _cvalue = utils._EngineCharScalarType.empty_ctypes(chrlen=32)
    _optype = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _handle,
        _optstr,
        _ivalue,
        _rvalue,
        _cvalue,
        _optype,
        _errbuf,
        _ierr,
        _optstr._length_,
        _cvalue._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _optstr_py = utils._EngineCharScalarType.to_py(_optstr)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'handle has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'handle is not a G22 handle.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, option was not recognized:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, option is not readable:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 121 and _ifmt == 121:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid instance identifier for option.',
                    'On entry, optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 122 and _ifmt == 122:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Numeric instance identifier is out of range.',
                    'On entry, instance identifier = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: {:s} <= instance identifier'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and instance identifier <= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 123 and _ifmt == 123:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, option cannot have an associated instance identifier. The supplied',
                    'instance identifier was ignored.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 124 and _ifmt == 124:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'option has multiple instances.',
                    'Information from the first instance has been returned.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ivalue_py = utils._EngineIntScalarType.to_py(_ivalue)
        _rvalue_py = utils._EngineFloat64ScalarType.to_py(_rvalue)
        _cvalue_py = utils._EngineCharScalarType.to_py(_cvalue)
        _optype_py = utils._EngineIntScalarType.to_py(_optype)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ivalue', 'rvalue', 'cvalue', 'optype'],
        )(
            _ivalue_py,
            _rvalue_py,
            _cvalue_py,
            _optype_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data
